# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.debug
    ~~~~~~~~~~~~~~

    The debug module provides the ability to print images as ASCII. It isn't a
    good ASCII representation like cacalib. This module is mostly intended for
    debugging purposes (hence the name.)

"""

from __future__ import print_function

import sys
import itertools

from fabulous import image


class DebugImage(image.Image):
    """Visualize optimization techniques used by :class:`Image`
    """

    def reduce(self, colors):
        need_reset = False
        line = ''
        for color, items in itertools.groupby(colors):
            if color is None:
                if need_reset:
                    line = line[:-1] + ">"
                    need_reset = False
                line += 'T' + (self.pad * len(list(items)))[1:]
            elif color == "EOL":
                if need_reset:
                    line = line[:-1] + ">"
                    need_reset = False
                    yield line.rstrip(' T')
                else:
                    yield line.rstrip(' T')
                line = ''
            else:
                need_reset = True
                line += '<' + (self.pad * len(list(items)))[1:]


def main():
    """I provide a command-line interface for this module
    """
    for imgpath in sys.argv[1:]:
        for line in DebugImage(imgpath):
            print(line)


if __name__ == '__main__':
    main()



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys


def printy(s):
    """Python 2/3 compatible print-like function"""
    if hasattr(s, 'as_utf8'):
        if hasattr(sys.stdout, 'buffer'):
            sys.stdout.buffer.write(s.as_utf8)
            sys.stdout.buffer.write(b"\n")
        else:
            sys.stdout.write(s.as_utf8)
            sys.stdout.write(b"\n")
    else:
        print(s)



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.xterm256
    ~~~~~~~~~~~~~~~~~

    The xterm256 module provides support for the 256 colors supported by xterm
    as well as quantizing 24-bit RGB color to xterm color ids.

    Color quantization may perform slowly depending on whether or not Fabulous
    is able to compile ``~/.xterm256.so`` on the fly. This is a tiny library
    that makes color quantization go much faster. The pure Python version of the
    algorithm is really slow because it's implemented as a brute force nearest
    neighbor over Euclidean distance search. Although an O(1) version of this
    algorithm exists with slightly less correctness. Your humble author simply
    hasn't had the time to implement it in this library.

"""

import logging


CUBE_STEPS = [0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF]
BASIC16 = ((0, 0, 0), (205, 0, 0), (0, 205, 0), (205, 205, 0),
           (0, 0, 238), (205, 0, 205), (0, 205, 205), (229, 229, 229),
           (127, 127, 127), (255, 0, 0), (0, 255, 0), (255, 255, 0),
           (92, 92, 255), (255, 0, 255), (0, 255, 255), (255, 255, 255))


def xterm_to_rgb(xcolor):
    """Convert xterm Color ID to an RGB value

    All 256 values are precalculated and stored in :data:`COLOR_TABLE`
    """
    assert 0 <= xcolor <= 255
    if xcolor < 16:
        # basic colors
        return BASIC16[xcolor]
    elif 16 <= xcolor <= 231:
        # color cube
        xcolor -= 16
        return (CUBE_STEPS[xcolor // 36 % 6],
                CUBE_STEPS[xcolor // 6 % 6],
                CUBE_STEPS[xcolor % 6])
    elif 232 <= xcolor <= 255:
        # gray tone
        c = 8 + (xcolor - 232) * 0x0A
        return (c, c, c)


COLOR_TABLE = [xterm_to_rgb(i) for i in range(256)]


def rgb_to_xterm(r, g, b):
    """Quantize RGB values to an xterm 256-color ID

    This works by envisioning the RGB values for all 256 xterm colors
    as 3D euclidean space and brute-force searching for the nearest
    neighbor.

    This is very slow.  If you're very lucky, :func:`compile_speedup`
    will replace this function automatically with routines in
    `_xterm256.c`.
    """
    if r < 5 and g < 5 and b < 5:
        return 16
    best_match = 0
    smallest_distance = 10000000000
    for c in range(16, 256):
        d = (COLOR_TABLE[c][0] - r) ** 2 + \
            (COLOR_TABLE[c][1] - g) ** 2 + \
            (COLOR_TABLE[c][2] - b) ** 2
        if d < smallest_distance:
            smallest_distance = d
            best_match = c
    return best_match


def compile_speedup():
    """Tries to compile/link the C version of this module

    Like it really makes a huge difference.  With a little bit of luck
    this should *just work* for you.

    You need:

    - Python >= 2.5 for ctypes library
    - gcc (``sudo apt-get install gcc``)

    """
    import os
    import ctypes
    from os.path import join, dirname, getmtime, exists, expanduser
    # library = join(dirname(__file__), '_xterm256.so')
    library = expanduser('~/.xterm256.so')
    sauce = join(dirname(__file__), '_xterm256.c')
    if not exists(library) or getmtime(sauce) > getmtime(library):
        build = "gcc -fPIC -shared -o %s %s" % (library, sauce)
        if (os.system(build + " >/dev/null 2>&1") != 0):
            raise OSError("GCC error")
    xterm256_c = ctypes.cdll.LoadLibrary(library)
    xterm256_c.init()
    def xterm_to_rgb(xcolor):
        res = xterm256_c.xterm_to_rgb_i(xcolor)
        return ((res >> 16) & 0xFF, (res >> 8) & 0xFF, res & 0xFF)
    return (xterm256_c.rgb_to_xterm, xterm_to_rgb)


try:
    (rgb_to_xterm, xterm_to_rgb) = compile_speedup()
except OSError:
    logging.debug("fabulous failed to compile xterm256 speedup code")



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.term
    ~~~~~~~~~~~~~

    Terminal abstraction layer.

    Provides standard capabilites to a variety of terminals. Support information
    is being worked on.

    .. code-block:: python

        import os
        os.stdout.write('spam' +
                        display('bright','yellow','white') +
                        'eggs' +
                        display('default') + os.linesep)

    **Warning:** on IPython setting sys.stdout to stdout will break readline

    **Caveat:** Failure to flush after ouput can cause weird ordering behaviour
    when writing to stdout and stderr simutaniously. This should fix the worst
    of it, but application developers should be warned not to rely on the state
    of things between call between one method call and another

"""

__all__ = ['display', 'stdin', 'stdout', 'stderr',
           'Term', 'UnixTerm', 'CursesTerm',
           'WinTerm', 'Win32Term', 'WinCTypesTerm']

import sys
import os
import re

try:
    unicode = unicode
except NameError:
    unicode = str
    basestring = (str, bytes)


# pylint: disable-msg=W0613
# pylint: disable-msg=W0102
# pylint: disable-msg=C0103
# pylint: disable-msg=W0142
# pylint: disable-msg=R0201
# pylint: disable-msg=W0511

class Term(object):
    """A file-like object which also supports terminal features.
    
    This is a base class for dumb terminals. It supports almost nothing.
    """
    
    def __init__(self, stream):
        """Under Construction (:P)
        
        When overriding this in subclasses, please call
        Term.__init__(self, stream) first.
        
        Class specific imports are being done in the contructors and
        references to the modules are stored as instance variables. This
        method is cleaner and works nicely with inheritance.
        """
        self.stream = stream
    
    def bell(self):
        """Causes the computer to beep
        
        Use sparingly, it is mainly
        to alert the user if something potentialy bad may be happening.
        """
        pass
    
    def display(self, codes=[], fg=None, bg=None):
        """Not for public consumption (yet)
        
        Just use display() and stdout.write() for now.
        
        run this at the beginning::
        
            (codes, fg, bg) = Magic.displayformat(codes, fg, bg)
        """
        pass
    
    def move(self, place, distance = 1):
        """Move cursor position
        
        The valid values for place are:
        
        up
            Move up a line.
        down
            Move to the next line. This also puts you at the beginning of
            the line.
        left
            Move one place to the left.
        right
            Move one place to the right.
        beginning of line
            Move to the beginning of the current line.
        beginning of screen
            Move to the beginning of the screen.
        """
        pass
    
    def clear(self, scope = 'screen'):
        """clears part of the screen
        
        The valid values for scope are:
        
        right
            Clears a single space directly to the right.
        left
            Clears a single space directly to the left.
        line
            Clears the current line.
        screen
            Clears the whole screen.
        beginning of line
            Clears from the current position to the beginning of the line
        end of line
            Clears from the current position to the end of the line
        end of screen
            Clears from the current position to the end of the screen
        
        N.b. this is not the same as deleting. After a place is cleared
        it should still be there, but with nothing in it. Also, this
        should not change the position of the cursor.
        """
        pass
    
    def get_size(self):
        """Get the width and height of the terminal.
        
        Returns either a tuple of two integers or None. If two integers are
        returned, the first one is the number of columns (or width) and the
        second value is the number of lines (or height). If None is returned,
        then the terminal does not support this feature. If you still need to
        have a value to fall back on (75, 25) is a fairly descent fallback.
        """
        return None
    
    def set_title(self, name):
        """Sets the title of the terminal
        """
        pass
    
    # methods below here are also methods of the file object
    
    def isatty(self):
        """Returns True if the terminal is a terminal
        
        This should always be True. If it's not somebody is being rather
        nauty.
        """
        return self.stream.isatty()
    
    def fileno(self):
        """Returns the stream's file descriptor as an integer"""
        return self.stream.fileno()
    
    # write-specific methods
    
    def write(self, text):
        """Parses text and prints proper output to the terminal
        
        This method will extract escape codes from the text and
        handle them as well as possible for whichever platform
        is being used. At the moment only the display escape codes
        are supported.
        """
        escape_parts = re.compile('\x01?\x1b\\[([0-9;]*)m\x02?')
        chunks = escape_parts.split(text)
        i = 0
        for chunk in chunks:
            if chunk != '':
                if i % 2 == 0:
                    self.stream.write(chunk)
                else:
                    c = chunk.split(';')
                    r = Magic.rdisplay(c)
                    self.display(**r) #see caveat 0
                self.flush()
            i += 1
    
    def writelines(self, sequence_of_strings):
        """Write out a sequence of strings
        
        Note that newlines are not added.  The sequence may be any iterable object
        producing strings. This is equivalent to calling write() for each string.
        """
        map(self.write, sequence_of_strings)
    
    def flush(self):
        """Ensure the text is ouput to the screen.
        
        The write() method will do this automatically, so only use this when
        using self.stream.write().
        """
        return self.stream.flush()
    
    # read-specific methods, they are in need of help
    
    def getch(self):
        """Don't use this yet
        
        It doesn't belong here but I haven't yet thought about a proper
        way to implement this feature and the features that will depend on
        it.
        """
        pass
    
    def raw_input(self, prompt):
        """Don't use this yet
        
        It doesn't belong here but I haven't yet thought about a proper
        way to implement this feature and the features that will depend on
        it.
        """
        return raw_input(prompt)
    
    def next(self):
        return self.stream.next()
    
    def readline(self, *args, **kwargs):
        return self.stream.readline(*args, **kwargs)
    def readlines(self, *args, **kwargs):
        return self.stream.readlines(*args, **kwargs)
    def read(self, *args, **kwargs):
        return self.stream.read(*args, **kwargs)
    
    # read-only properties
    @property
    def mode(self):
        return self.stream.mode
    @property
    def newlines(self):
        return self.stream.newlines
    @property
    def encoding(self):
        return self.stream.encoding
    @property
    def softspace(self):
        return self.stream.softspace
    @property
    def name(self):
        return self.stream.name

class UnixTerm(Term):
    
    def __init__(self, stream):
        import termios
        import tty
        self.termios = termios
        self.tty = tty
        Term.__init__(self, stream)
    
    def getch(self):
        """Don't use this yet
        
        It doesn't belong here but I haven't yet thought about a proper
        way to implement this feature and the features that will depend on
        it.
        """
        return NotImplemented
        fno = stdout.fileno()
        mode = self.termios.tcgetattr(fno)
        try:
            self.tty.setraw(fno, self.termios.TCSANOW)
            ch = self.read(1)
        finally:
            self.termios.tcsetattr(fno, self.termios.TCSANOW, mode)
        return ch

class CursesTerm(UnixTerm):
    
    def __init__(self, stream):
        import curses
        self.curses = curses
        UnixTerm.__init__(self, stream)
        if not sys.stdout.isatty(): return
        self.curses.setupterm()
    
    def bell(self):
        self.stream.write(self._get_cap('bel'))
    
    def display(self, codes=[], fg=None, bg=None):
        """Displays the codes using ANSI escapes
        """
        codes, fg, bg = Magic.displayformat(codes, fg, bg)
        self.stream.write(Magic.display(codes, fg, bg))
        self.flush()
        
    def move(self, place, distance = 1):
        """see doc in Term class"""
        for d in range(distance):
            self.stream.write(self._get_cap('move '+place))
        self.flush()
    
    def clear(self, scope = 'screen'):
        """see doc in Term class"""
        if scope == 'line':
            self.clear('beginning of line')
            self.clear('end of line')
        else: self.stream.write(self._get_cap('clear '+scope))
        self.flush()
    
    def get_size(self):
        """see doc in Term class"""
        self.curses.setupterm()
        return self.curses.tigetnum('cols'), self.curses.tigetnum('lines')
    
    def set_title(self, name):
        self.write(Magic.OSC + '0;'+str(name) + "\x07") 
    
    def _get_cap(self, cap):
        strcaps = {
       'move up':'cuu1', 'move down':'cud1', 
           'move left':'cub1', 'move right':'cuf1',
           'move beginning of line':'cr', 'move beginning of screen':'home',
       'clear beginning of line':'el1','clear end of line':'el',
           'clear screen':'clear', 'clear end of screen':'ed',
           'clear left':'kbs','clear right':'dch1',
       'delete line':'dl1',
       'bell':'bel'}
        if cap in ('cols','lines'):
            self.curses.setupterm()
            c = self.curses.tigetnum(cap)
            if c > 0: return c
        elif strcaps.has_key(cap):
            c = self.curses.tigetstr(strcaps[cap])
            if c != '': return c
        raise ValueError("capability '%s' not supported" % cap)
    

class WinTerm(Term):
    """Windows version of terminal control
    
    This class should not be used by itself, use either Win32Terminal or 
    WinCTypesTerminal classes that subclasses of this class.
    
    This class makes extensive use of the Windows API
    
    The official documentation for the API is on MSDN (look for 'console
    functions')
    """
    
    # TODO: is there a better way to get this?
    
    STD_INPUT_HANDLE = -10
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12
    
    # These contants are defined in PyWin32
    # You can combine the values by doing a bitwise or (|)
    # for example FG_BLUE | FG_RED would give magenta (0x05)
    #
    # these contants are just numbers, It's most useful to think of
    # them in binary
    FG_BLUE = 1 << 0
    FG_GREEN = 1 << 1
    FG_RED = 1 << 2
    FG_INTENSITY = 1 << 3
    BG_BLUE = 1 << 4
    BG_GREEN = 1 << 5
    BG_RED = 1 << 6
    BG_INTENSITY = 1 << 7
    FG_ALL = FG_BLUE | FG_GREEN | FG_RED
    BG_ALL = BG_BLUE | BG_GREEN | BG_RED
    
    # there are also these codes, but according to tcsh's win32/console.c:
    # COMMON_LVB_REVERSE_VIDEO is buggy, so I'm staying away from it. Future
    # versions should implement COMMON_LVB_UNDERSCORE.
    # COMMON_LVB_REVERSE_VIDEO = 0x4000
    # COMMON_LVB_UNDERSCORE      0x8000
    
    FG = {
    'black': 0,
    'red': FG_RED,
    'green': FG_GREEN,
    'yellow': FG_GREEN | FG_RED,
    'blue': FG_BLUE,
    'magenta': FG_BLUE | FG_RED,
    'cyan': FG_BLUE | FG_GREEN,
    'white': FG_BLUE | FG_GREEN | FG_RED,
    }
    BG = {
    'black':0,
    'red':BG_RED,
    'green':BG_GREEN,
    'yellow':BG_GREEN | BG_RED,
    'blue':BG_BLUE,
    'magenta':BG_BLUE | BG_RED,
    'cyan':BG_BLUE | BG_GREEN,
    'white':BG_BLUE | BG_GREEN | BG_RED,
    }
    
    default_attributes = None
    hidden_output = False
    reverse_output = False
    reverse_input = False
    dim_output = False
    real_fg = None
    
    def __init__(self, stream):
        import msvcrt
        self.msvcrt = msvcrt
        Term.__init__(self, stream)
        self._stdout_handle = self._get_std_handle(self.STD_OUTPUT_HANDLE)
        self._stderr_handle = self._get_std_handle(self.STD_ERROR_HANDLE)
        self.default_attributes = self._get_console_info()['attributes']
        self.real_fg = self.default_attributes & 0x7
    
    def display(self, codes=[], fg=None, bg=None):
        """Displays codes using Windows kernel calls
        """
        codes, fg, bg = Magic.displayformat(codes, fg, bg)
        color = 0
        for c in codes:
            try:
                f = getattr(self, '_display_' + c)
                out = f()
                if out: color |= out
            except AttributeError:
                pass
        cfg, cfgi, cbg, cbgi = self._split_attributes(
                          self._get_console_info()['attributes'])
        if self.reverse_input:
            cfg, cbg = (cbg // 0x10), (cfg * 0x10)
            cfgi, cbgi = (cbgi // 0x10), (cfgi * 0x10)
        if fg != None:
            color |= self.FG[fg]
            self.real_fg = self.FG[fg]
        else: color |= cfg
        if bg != None:
            color |= self.BG[bg]
        else: color |= cbg
        color |= (cfgi | cbgi)
        fg, fgi, bg, bgi = self._split_attributes(color)
        if self.dim_output:
            # intense black
            fg = 0
            fgi = self.FG_INTENSITY
        if self.reverse_output:
            fg, bg = (bg // 0x10), (fg * 0x10)
            fgi, bgi = (bgi // 0x10), (fgi * 0x10)
            self.reverse_input = True
        if self.hidden_output:
            fg = (bg // 0x10)
            fgi = (bgi // 0x10)
        self._set_attributes(fg | fgi | bg | bgi)
    
    def get_size(self):
        """see doc in Term class"""
        attr = self._get_console_info()
        cols = attr['window']['right'] - attr['window']['left'] + 1
        lines = attr['window']['bottom'] - attr['window']['top'] + 1
        return cols, lines
    
    def _get_std_handle(self, handleno):
        """Returns a handle from GetStdHandle
        
        handleno is one of:
        * self.STD_OUTPUT_HANDLE for stdout
        * self.STD_ERROR_HANDLE for stderr
        """
        #TODO: is NotImplemented the proper way to do this?
        return NotImplemented
    
    def _get_console_info(self):
        """Get information from GetConsoleScreenBufferInfo
        
        Returns a dictionary with the following keys::
        
            max size
            position
            window
            attributes
            size
        
        Note: the y part of size is misleading
        """
        return NotImplemented
    
    def _clear_console(self, length, start):
        """Clears a part of the console
        
        Has a similar effect as writing out spaces.
        
        length: int length of cleared section
        start : tuple of x and y coords to start at
        """
        return NotImplemented
    
    def _set_attributes(self, code):
        """ Set console attributes with `code`
        
        Not implemented here. To be implemented by subclasses.
        """
        return NotImplemented
    
    def _split_attributes(self, attrs):
        """Spilt attribute code
        
        Takes an attribute code and returns a tuple containing
        foreground (fg), foreground intensity (fgi), background (bg), and
        background intensity (bgi)
        
        Attributes can be joined using ``fg | fgi | bg | bgi``
        """
        fg = attrs & self.FG_ALL
        fgi = attrs & self.FG_INTENSITY
        bg = attrs & self.BG_ALL
        bgi = attrs & self.BG_INTENSITY
        return fg, fgi, bg, bgi
    
    def _undim(self):
        self.dim_output = False
        if self.reverse_input:
            a = self._get_console_info()['attributes'] & 0x8f
            self._set_attributes( (self.real_fg * 0x10) | a)
        else:
            a = self._get_console_info()['attributes'] & 0xf8
            self._set_attributes(self.real_fg | a)
    
    def _display_default(self):
        self.hidden_output = False
        self.reverse_output = False
        self.reverse_input = False
        self.dim_output = False
        self.real_fg = self.default_attributes & 0x7
        self._set_attributes(self.default_attributes)

    def _display_bright(self):
        self._undim()
        return self.FG_INTENSITY
    
    def _display_dim(self):
        self.dim_output = True
    
    def _display_reverse(self):
        self.reverse_output = True
    
    def _display_hidden(self):
        self.hidden_output = True
    
    def _get_position(self):
        """Set the cursor's current position
        
        Returns a tuple in the form (x, y)
        """
        pos = self._get_console_info()['position']
        return pos['x'], pos['y']
    
    
    def _set_position(self, coord):
        """Set the cursor's position
        
        coord is a tuple in the form (x, y)
        """
        return NotImplemented
    
    def move(self, place, distance = 1):
        """see doc in Term class"""
        x, y = self._get_position()
        if place == 'up':
            y -= distance
        elif place == 'down':
            for i in range(distance): print
            nx, ny = self._get_position()
            y = ny
            self.move('beginning of line')
        elif place == 'left':
            x -= distance
        elif place == 'right':
            x += distance
        elif place == 'beginning of line':
            x = 0
        elif place == 'beginning of screen':
            x = 0
            y = self._get_console_info()['window']['top']
        else:
            raise ValueError("invalid place to move")
        self._set_position((x, y))
    
    def clear(self, scope = 'screen'):
        """see doc in Term class
        
        According to http://support.microsoft.com/kb/99261 the best way
        to clear the console is to write out empty spaces
        """
        #TODO: clear attributes too
        if scope == 'screen':
            bos = (0, self._get_console_info()['window']['top'])
            cols, lines = self.get_size()
            length = cols * lines
            self._clear_console(length, bos)
            self.move('beginning of screen')
        elif scope == ' beginning of line':
            pass
        elif scope == 'end of line':
            curx, cury = self._get_position()
            cols, lines = self.get_size()
            coord = (curx, cury)
            length = cols - curx
            self._clear_console(length, coord)
        elif scope == 'end of screen':
            curx, cury = self._get_position()
            coord = (curx, cury)
            cols, lines = self.get_size()
            length = (lines - cury) * cols - curx
            self._clear_console(length, coord)
        elif scope == 'line':
            curx, cury = self._get_position()
            coord = (0, cury)
            cols, lines = self.get_size()
            self._clear_console(cols, coord)
            self._set_position((curx, cury))
        elif scope == 'left':
            self.move('left')
            self.write(' ')
        elif scope == 'right':
            self.write(' ')
            self.move('left')
        else:
            raise ValueError("invalid scope to clear")
        
    def getch(self):
        """Don't use this yet
        
        It doesn't belong here but I haven't yet thought about a proper
        way to implement this feature and the features that will depend on
        it.
        """
        return NotImplemented
        return self.msvcrt.getch()
    
    def bell(self):
        self.stream.write('\x07')

class Win32Term(WinTerm):
    """PyWin32 version of Windows terminal control.
    
    Uses the PyWin32 Libraries <http://sourceforge.net/projects/pywin32/>.
    
    ActiveState has good documentation for them:
    
    Main page:
    http://aspn.activestate.com/ASPN/docs/ActivePython/2.4/pywin32/PyWin32.html
    Console related objects and methods:
    http://aspn.activestate.com/ASPN/docs/ActivePython/2.4/pywin32/PyConsoleScreenBuffer.html
    """
    
    def __init__(self, stream):
        import win32console
        self.win32console = win32console
        WinTerm.__init__(self, stream)
    
    def set_title(self, name):
        return self.win32console.SetConsoleTitle(name)
    
    def _get_console_info(self):
        # example output from GetConsoleScreenBufferInfo
        # {'MaximumWindowSize': PyCOORDType(X=80,Y=82),
        # 'CursorPosition': PyCOORDType(X=0,Y=6),
        # 'Window': PySMALL_RECTType(Left=0,Top=0,Right=79,Bottom=24),
        # 'Attributes': 7,
        # 'Size': PyCOORDType(X=80,Y=300)}
        attrs = self._stdout_handle.GetConsoleScreenBufferInfo()
        return {'max size': self._pyCoord_dict(attrs['MaximumWindowSize']),
                'position': self._pyCoord_dict(attrs['CursorPosition']),
                'window': self._pySMALL_RECTType_dict(attrs['Window']),
                'attributes': attrs['Attributes'],
                # y part of size value is misleading
                'size': self._pyCoord_dict(attrs['Size']) }
    
    def _get_std_handle(self, handle):
        return self.win32console.GetStdHandle(handle)
    
    def _get_title(self):
        return self.win32console.GetConsoleTitle()
    
    def _set_attributes(self, attr):
        self._stdout_handle.SetConsoleTextAttribute(attr)
    
    def _set_position(self, coord):
        coord = self.win32console.PyCOORDType(coord[0], coord[1])
        self._stdout_handle.SetConsoleCursorPosition(coord)
        
    def _clear_console(self, length, start):
        # length: int
        # start : tuple of x and y coords
        char = unicode(' ')
        coord = self.win32console.PyCOORDType(start[0], start[1])
        # char is unicode
        self._stdout_handle.FillConsoleOutputCharacter(
           char, length, coord)
    
    def _pyCoord_dict(self, coord):
        return { 'x': coord.X, 'y': coord.Y}
    
    def _pySMALL_RECTType_dict(self, rect):
        return { 'left': rect.Left, 'top': rect.Top,
                'right': rect.Right, 'bottom': rect.Bottom}

class WinCTypesTerm(WinTerm):
    """CTypes version of Windows terminal control.
    
    It requires the CTypes libraries <http://sourceforge.net/projects/ctypes/>
    
    As of Python 2.5, CTypes is included in Python by default. User's of
    previous version of Python will have to install it if they what to use
    this.
    """
    
    def __init__(self, stream):
        import ctypes
        self.ctypes = ctypes
        WinTerm.__init__(self, stream)
    
    def set_title(self, name):
        self.ctypes.windll.kernel32.SetConsoleTitleA(name)
    
    def _get_console_info(self):
        # From IPython's winconsole.py, by Alexander Belchenko
        import struct
        csbi = self.ctypes.create_string_buffer(22)
        res = self.ctypes.windll.kernel32.GetConsoleScreenBufferInfo(
                                     self._stdout_handle, csbi)
        (bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx,
         maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
        return {'max size': {'x':maxx, 'y':maxy },
                'position': {'x':curx, 'y':cury },
                'window': {'left': left, 'top': top,
                           'right': right, 'bottom': bottom},
                'attributes': wattr,
                # y part of size value is misleading
                'size': {'x':maxx, 'y':maxy } }
    
    def _get_std_handle(self, handle):
        return self.ctypes.windll.kernel32.GetStdHandle(handle)
    
    def _get_title(self):
        """According to http://support.microsoft.com/kb/124103 the buffer
        size is 1024
        
        Does not support unicode, only ANSI"""
        #TODO: unicode support
        strbuffer = self.ctypes.create_string_buffer(1024)
        size = self.ctypes.c_short(1024)
        #unicode versions are (Get|Set)ConsolTitleW
        self.ctypes.windll.kernel32.GetConsoleTitleA(strbuffer, size)
        return strbuffer.value
    
    def _set_attributes(self, attr):
        self.ctypes.windll.kernel32.SetConsoleTextAttribute(
                        self._stdout_handle, attr)
    
    def _set_position(self, coord):
        coord = self._get_coord(coord)
        self.ctypes.windll.kernel32.SetConsoleCursorPosition(
                        self._stdout_handle, coord)
        
    def _clear_console(self, length, start):
        # length: int
        # start : tuple of x and y coords
        char = self.ctypes.c_char(' ')
        coord = self._get_coord(start)
        charswritten = self.ctypes.c_int()
        clength = self.ctypes.c_int(length)
        self.ctypes.windll.kernel32.FillConsoleOutputCharacterA(
           self._stdout_handle, char, clength, coord, charswritten)
    
    def _get_coord(self, coord):
        """ It's a hack, see fixcoord in pyreadline's console.py (revision 
        1289)
        """
        x, y = coord
        return self.ctypes.c_int(y << 16 | x)

class Magic(object):
    """Special codes and what not
    
    Don't use these alone
    see http://vt100.net/docs/vt100-ug/chapter3.html
    
    Based on the ANSI X3.64 standard. See http://en.wikipedia.org/wiki/ANSI_X3.64
    """
    ESCAPE = '\x1b'
    CSI = ESCAPE +'['
    OSC = ESCAPE +']'
    # see the reset method
    RESET = ESCAPE + 'c'
    
    
    # pylint: disable-msg=E0602
    DISPLAY = {'default':0, 'bright':1, 'dim':2, 'underline':4, 'blink':5,
             'reverse':7, 'hidden':8 }
    rDISPLAY = dict( (v, k) for k, v in DISPLAY.items())
    # Yellow is a bit weird, xterm and rxvt display dark yellow, while linux
    # and Windows display a more brown-ish color. Bright yellow is always
    # yellow. Order is important here
    COLORS = { 'black':0, 'red':1, 'green':2, 'yellow':3, 'blue':4, 'magenta':5,
               'cyan':6, 'white':7 }
    rCOLORS = dict( (v, k) for k, v in COLORS.items())
    # pylint: enable-msg=E0602
    # TODO: setf from curses uses the colors in a different order for 
    
    @staticmethod
    def displayformat(codes=[], fg=None, bg=None):
        """Makes sure all arguments are valid"""
        if isinstance(codes, basestring):
            codes = [codes]
        else:
            codes = list(codes)
        for code in codes:
            if code not in Magic.DISPLAY.keys():
                raise ValueError("'%s' not a valid display value" % code)
        for color in (fg, bg):
            if color != None:
                if color not in Magic.COLORS.keys():
                    raise ValueError("'%s' not a valid color" % color)
        return [codes, fg, bg]
    
    @staticmethod
    def rdisplay(codes):
        """Reads a list of codes and generates dict
        
        >>> Magic.rdisplay([])
        {}
        >>> result = Magic.rdisplay([1,2,34,46])
        >>> sorted(result.keys())
        ['bg', 'codes', 'fg']
        >>> sorted(result['codes'])
        ['bright', 'dim']
        >>> result['bg']
        'cyan'
        >>> result['fg']
        'blue'
        """
        dcodes = []
        fg = bg = None
        for code in codes:
            code = int(code)
            offset = code // 10
            decimal = code % 10
            if offset == 3 and decimal in Magic.COLORS.values(): fg = decimal
            elif offset == 4 and decimal in Magic.COLORS.values(): bg = decimal
            elif code in Magic.DISPLAY.values(): dcodes.append(code)
            else: pass # drop unhandled values
        r = {}
        if len(codes): r['codes'] = [Magic.rDISPLAY[c] for c in dcodes]
        if fg != None: r['fg'] = Magic.rCOLORS[fg]
        if bg != None: r['bg'] = Magic.rCOLORS[bg]
        return r

    @staticmethod
    def display(codes=[], fg=None, bg=None):
        codes, fg, bg = Magic.displayformat(codes, fg, bg)
        codes = [str(Magic.DISPLAY[code]) for code in codes]
        if fg != None: codes.append(str(30 + Magic.COLORS[fg]))
        if bg != None: codes.append(str(40 + Magic.COLORS[bg]))
        return Magic.CSI + ";".join(codes) + 'm'

def display(codes=[], fg=None, bg=None):
    """Returns an ANSI display code. This is useful when writing to an Term
        
    codes
        A list containing strings. The strings should one of the keys in
        ``Magic.DISPLAY``. It can also be just a single string.
    fg, bg
        A string. Explicitly for setting the foreground or background. Use
        one of the keys in ``Magic.COLORS``.
        
    .. code-block:: python
    
        # give bright blue foreground and white background with underline
        display(('bright','underline'),'blue','white')
        # gives a blue foreground
        display(fg='blue')
        # resets the color to the default.
        display('default')
    
    Avoid using black or white. Depending on the situation the default
    background/foreground is normally black or white, but it's hard to
    tell which. Bare terminals are normally white on black, but virtual
    terminals run from X or another GUI system are often black on white.
    This can lead to unpredicatble results. If you want reversed
    colours, use the 'reverse' code, and if you want to set the
    colors back to their original colors, use the 'default' code.
    
    Also, be prudent with your use of 'hidden' and 'blink'. Several terminals
    do not support them (and for good reason too), they can be really
    annoying and make reading difficult.
    """
    return Magic.display(codes, fg, bg)

# try to use the Windows method first because their are some terminals on
# MS Windows that support both the Windows and curses methods, but their
# curses implementations are buggy.

def _get_terms():
    terms = None
    if 'win32' in sys.platform or 'cygwin' == sys.platform:
        terms = _get_term(WinCTypesTerm) or _get_term(Win32Term)
    if not terms:
        terms = (_get_term(CursesTerm) or 
                 _get_term(UnixTerm) or 
                 _get_term(Term))
    return terms

def _get_term(termclass):
    try:
        return (termclass(sys.stdin), termclass(sys.stdout),
                termclass(sys.stderr))
    except ImportError: return None

stdin, stdout, stderr = _get_terms()



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fabulous import text, utils, image, debug, xterm256
from fabulous.color import *
from fabulous.compatibility import printy
import fabulous

try:
    input = raw_input
except NameError:
    pass


def wait():
    input("\nPress " + bold("enter") + " for more fun... ")
    printy("")


def demo_image():
    section("Semi-Transparent PNG")
    imp = "  from fabulous import image\n  "
    printy(bold(imp + 'print image.Image("balls.png")\n'))

    balls = 'balls.png'
    fabdir = os.path.dirname(fabulous.__file__)

    for fn in ['balls.png',
               'fabulous/balls.png',
               os.path.join(fabdir, 'balls.png')]:
        if os.path.exists(fn):
            balls = fn
            break

    if not os.path.exists(balls):
        import urllib
        ugh = urllib.urlopen('http://lobstertech.com/media/img/balls.png')
        open('balls.png', 'w').write(ugh.read())
        balls = 'balls.png'

    for line in image.Image(balls):
        printy(line)
    wait()

    section("Yes the output is optimized (JELLY-FISH)")
    imp = "  from fabulous import debug\n  "
    printy(bold(imp + 'print debug.DebugImage("balls.png")\n'))
    for line in debug.DebugImage(balls):
        printy(line)
    wait()


def demo_text():
    section('Fabulous Text Rendering')
    imp = "  from fabulous import text\n  "
    printy(bold(imp + 'print text.Text("Fabulous", shadow=True, skew=5)\n'))
    printy(text.Text("Fabulous", shadow=True, skew=5))
    wait()


def demo_color_4bit():
    section("Fabulous 4-Bit Colors")

    printy(("style(...): " +
            bold("bold") +" "+
            underline("underline") +" "+
            flip("flip") +
            " (YMMV: " + italic("italic") +" "+
            underline2("underline2") +" "+
            strike("strike") +" "+
            blink("blink") + ")\n"))

    printy(("color(...)           " +
            black("black") +" "+
            red("red") +" "+
            green("green") +" "+
            yellow("yellow") +" "+
            blue("blue") +" "+
            magenta("magenta") +" "+
            cyan("cyan") +" "+
            white("white")))

    printy(("bold(color(...))     " +
            bold(black("black") +" "+
                 red("red") +" "+
                 green("green") +" "+
                 yellow("yellow") +" "+
                 blue("blue") +" "+
                 magenta("magenta") +" "+
                 cyan("cyan") +" "+
                 white("white"))))

    printy(plain(
        'highlight_color(...) ',
        highlight_black('black'), ' ', highlight_red('red'), ' ',
        highlight_green('green'), ' ', highlight_yellow('yellow'), ' ',
        highlight_blue('blue'), ' ', highlight_magenta('magenta'), ' ',
        highlight_cyan('cyan'), ' ', highlight_white('white')))

    printy(("bold(color_bg(...))  " +
            bold(black_bg("black") +" "+
                 red_bg("red") +" "+
                 green_bg("green") +" "+
                 yellow_bg("yellow") +" "+
                 blue_bg("blue") +" "+
                 magenta_bg("magenta") +" "+
                 cyan_bg("cyan") +" "+
                 white_bg("white"))))

    wait()


def demo_color_8bit():
    section("Fabulous 8-Bit Colors")

    for code in ["bold(fg256('red', ' lorem ipsum '))",
                 "bold(bg256('#ff0000', ' lorem ipsum '))",
                 "highlight256((255, 0, 0), ' lorem ipsum ')",
                 "highlight256('#09a', ' lorem ipsum ')",
                 "highlight256('green', ' lorem ipsum ')",
                 "highlight256('magenta', ' lorem ipsum ')",
                 "highlight256('indigo', ' lorem ipsum ')",
                 "highlight256('orange', ' lorem ipsum ')",
                 "highlight256('orangered', ' lorem ipsum ')"]:
        printy("%-42s %s" % (code, eval(code)))
    printy('')

    # grayscales
    line = " "
    for xc in range(232, 256):
        line += bg256(xc, '  ')
    printy(line)
    line = " "
    for xc in range(232, 256)[::-1]:
        line += bg256(xc, '  ')
    printy(line)
    printy('')

    cube_color = lambda x,y,z: 16 + x + y*6 + z*6*6
    for y in range(6):
        line = " "
        for z in range(6):
            for x in range(6):
                line += bg256(cube_color(x, y, z), '  ')
            line += " "
        printy(line)

    wait()


def full_chart():
    # grayscales
    line = " "
    for xc in range(232, 256):
        line += bg256(xc, '  ')
    printy(line)
    line = " "
    for xc in range(232, 256)[::-1]:
        line += bg256(xc, '  ')
    printy(line)
    printy('')

    # cube
    printy("")
    cube_color = lambda x,y,z: 16 + x + y*6 + z*6*6
    for y in range(6):
        line = " "
        for z in range(6):
            for x in range(6):
                line += bg256(cube_color(x, y, z), '  ')
            line += " "
        printy(line)
    printy("")

    def f(xc):
        s = highlight256(xc, "color %03d" % (xc))
        rgb = xterm256.xterm_to_rgb(xc)
        rgbs = ' (%3d, %3d, %3d)' % rgb
        if rgb[0] == rgb[1] == rgb[2]:
            s += bold(rgbs)
        else:
            s += rgbs
        s += ' (%08d, %08d, %08d)' % tuple([int(bin(n)[2:]) for n in rgb])
        return s

    def l(c1, c2):
        c1, c2 = f(c1), f(c2)
        assert len(c1) == len(c2)
        half = width // 2
        assert half > len(c1)
        pad = " " * ((width // 2 - len(c1)) // 2)
        printy("%(pad)s%(c1)s%(pad)s%(pad)s%(c2)s" % {
            'pad': pad, 'c1': c1, 'c2': c2})

    width = utils.term.width
    for z1, z2 in zip((0, 2, 4), (1, 3, 5)):
        for y1, y2 in zip(range(6), range(6)):
            for x1, x2 in zip(range(6), range(6)):
                l(cube_color(x1, y1, z1), cube_color(x2, y2, z2))
        printy("")


def main():
    # full_chart()
    demo_color_4bit()
    demo_color_8bit()
    demo_text()
    demo_image()


if __name__ == '__main__':
    main()



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.image
    ~~~~~~~~~~~~~~

    The image module makes it possible to print images to the terminal.

    This module is available as a command line tool::

        jart@compy:~$ fabulous-image foo.png
        jart@compy:~$ python -m fabulous.image foo.png

"""

import sys
import itertools

from fabulous import utils, xterm256, grapefruit
from fabulous.compatibility import printy


class Image(object):
    """Printing image files to a terminal

    I use :mod:`PIL` to turn your image file into a bitmap, resize it
    so it'll fit inside your terminal, and implement methods so I can
    behave like a string or iterable.

    When resizing, I'll assume that a single character on the terminal
    display is one pixel wide and two pixels tall.  For most fonts
    this is the best way to preserve the aspect ratio of your image.

    All colors are are quantized by :mod:`fabulous.xterm256` to the
    256 colors supported by modern terminals.  When quantizing
    semi-transparant pixels (common in text or PNG files) I'll ask
    :class:`TerminalInfo` for the background color I should use to
    solidify the color.  Fully transparent pixels will be rendered as
    a blank space without color so we don't need to mix in a
    background color.

    I also put a lot of work into optimizing the output line-by-line
    so it needs as few ANSI escape sequences as possible.  If your
    terminal is kinda slow, you're gonna want to buy me a drink ;) You
    can use :class:`DebugImage` to visualize these optimizations.

    The generated output will only include spaces with different
    background colors.  In the future routines will be provided to
    overlay text on top of these images.

    """

    pad = ' '

    def __init__(self, path, width=None):
        utils.pil_check()
        from PIL import Image as PillsPillsPills
        self.img = PillsPillsPills.open(path)
        # when reading pixels, gifs will return colors corresponding
        # to a palette if we don't do this :\
        self.img = self.img.convert("RGBA")
        self.resize(width)

    def __iter__(self):
        """I allow Image to behave as an iterable

        By using me with a for loop, you can use each line as they're
        created.  When printing a large image, this helps you not have
        to wait for the whole thing to be converted.

        :return: Yields lines of text (without line end character)
        """
        # strip out blank lines
        for line in self.reduce(self.convert()):
            if line.strip():
                yield line
        yield ""

    def __str__(self):
        """I return the entire image as one big string

        Unlike the iteration approach, you have to wait for the entire
        image to be converted.

        :return: String containing all lines joined together.
        """
        return "\n".join(self)

    @property
    def size(self):
        """Returns size of image
        """
        return self.img.size

    def resize(self, width=None):
        """Resizes image to fit inside terminal

        Called by the constructor automatically.
        """
        (iw, ih) = self.size
        if width is None:
            width = min(iw, utils.term.width)
        elif (sys.version_info >= (3, 0) and isinstance(width, str)) or (sys.version_info < (3, 0) and isinstance(width, basestring)):
            percents = dict([(pct, '%s%%' % (pct)) for pct in range(101)])
            width = percents[width]
        height = int(float(ih) * (float(width) / float(iw)))
        height //= 2
        self.img = self.img.resize((width, height))

    def reduce(self, colors):
        """Converts color codes into optimized text

        This optimizer works by merging adjacent colors so we don't
        have to repeat the same escape codes for each pixel.  There is
        no loss of information.

        :param colors: Iterable yielding an xterm color code for each
                       pixel, None to indicate a transparent pixel, or
                       ``'EOL'`` to indicate th end of a line.

        :return: Yields lines of optimized text.

        """
        need_reset = False
        line = []
        for color, items in itertools.groupby(colors):
            if color is None:
                if need_reset:
                    line.append("\x1b[49m")
                    need_reset = False
                line.append(self.pad * len(list(items)))
            elif color == "EOL":
                if need_reset:
                    line.append("\x1b[49m")
                    need_reset = False
                    yield "".join(line)
                else:
                    line.pop()
                    yield "".join(line)
                line = []
            else:
                need_reset = True
                line.append("\x1b[48;5;%dm%s" % (
                    color, self.pad * len(list(items))))

    def convert(self):
        """Yields xterm color codes for each pixel in image
        """
        (width, height) = self.img.size
        bgcolor = utils.term.bgcolor
        self.img.load()
        for y in range(height):
            for x in range(width):
                rgba = self.img.getpixel((x, y))
                if len(rgba) == 4 and rgba[3] == 0:
                    yield None
                elif len(rgba) == 3 or rgba[3] == 255:
                    yield xterm256.rgb_to_xterm(*rgba[:3])
                else:
                    color = grapefruit.Color.NewFromRgb(
                        *[c / 255.0 for c in rgba])
                    rgba = grapefruit.Color.AlphaBlend(color, bgcolor).rgb
                    yield xterm256.rgb_to_xterm(
                        *[int(c * 255.0) for c in rgba])
            yield "EOL"


def main():
    """Main function for :command:`fabulous-image`."""
    import optparse
    parser = optparse.OptionParser()
    parser.add_option(
        "-w", "--width", dest="width", type="int", default=None,
        help=("Width of printed image in characters.  Default: %default"))
    (options, args) = parser.parse_args(args=sys.argv[1:])
    for imgpath in args:
        for line in Image(imgpath, options.width):
            printy(line)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.color
    ~~~~~~~~~~~~~~

    The color module provides an object-oriented abstraction for stylized text
    inside the terminal. This includes things like bold text, blinking text,
    4-bit ANSI colors, 8-bit xterm256 colors, and 24-bit "truecolor" colors.

"""

import sys
import functools

from fabulous.compatibility import printy
from fabulous import utils, xterm256, grapefruit

try:
    unicode = unicode
except NameError:
    unicode = str
    basestring = (str, bytes)


OVERLINE = u'\u203e'


def esc(*codes):
    """Produces an ANSI escape code string from a list of integers

    This is a low level function that is abstracted by the other functions and
    classes in this module.

    """
    return "\x1b[%sm" % (";".join([str(c) for c in codes]))


class ColorString(object):
    r"""Abstract base class for stylized string-like objects.

    Subclasses make it possible to compose stylized text::

        >>> str(red("hello"))
        '\x1b[31mhello\x1b[39m'
        >>> str(bold(red("hello")))
        '\x1b[1m\x1b[31mhello\x1b[39m\x1b[22m'
        >>> str(plain("hello ", bold("world")))
        'hello \x1b[1mworld\x1b[22m'

    These objects also provide string length without taking into consideration
    the ANSI escape codes::

        >>> len(red("hello"))
        5
        >>> len(str(red("hello")))
        15
        >>> len(bold(red("hello")))
        5
        >>> len(bold("hello ", red("world")))
        11

    If you have the wcwidth module installed, it will be used for computing lengths::



    """
    sep = ""
    fmt = "%s"

    def __init__(self, *items):
        self.items = items

    def __str__(self):
        return self.fmt % (self.sep.join([unicode(s) for s in self.items]))

    def __repr__(self):
        return repr(unicode(self))


    if sys.version_info[0] > 2:
        def __len__(self):
            try:
                import wcwidth
                # We use:
                # * wcwidth.wcswidth(item) to find the length of strs
                # * len(str(item)) to find the length of a bytes object
                # * len(item) for everything else.
                return sum([ wcwidth.wcswidth(item) if isinstance(item, str) else len(str(item)) if isinstance(item, bytes) else len(item) for item in self.items ])
            except ModuleNotFoundError:
                return sum([len(str(item)) if isinstance(item, bytes) else len(item) for item in self.items])
    else:
        def __len__(self):
            return sum([len(item) for item in self.items])

    def __add__(self, cs):
        if not isinstance(cs, (basestring, ColorString)):
            msg = "Concatenatation failed: %r + %r (Not a ColorString or str)"
            raise TypeError(msg % (type(cs), type(self)))
        return ColorString(self, cs)

    def __radd__(self, cs):
        if not isinstance(cs, (basestring, ColorString)):
            msg = "Concatenatation failed: %r + %r (Not a ColorString or str)"
            raise TypeError(msg % (type(self), type(cs)))
        return ColorString(cs, self)

    @property
    def as_utf8(self):
        """A more readable way to say ``unicode(color).encode('utf8')``
        """
        return unicode(self).encode('utf8')
    def join(self, iterable):
        """
        This works just like `str.join()`, but for ColorStrings!

        For example:

            >>> from fabulous.color import *
            >>> l = [
            ...     fg256("green", "napster good"),
            ...     fg256("red", "fire bad"),
            ... ]
            >>> print(plain(" ").join(l))

        """
        ret = None
        for x in iterable:
            if ret is None:
                ret = x
            else:
                ret += self
                ret += x
        return ret


class ColorString256(ColorString):
    r"""Base class for 256-color stylized string-like objects.

    See the :class:`.fg256`, :class:`.bg256`, :class:`.highlight256`, and
    :class:`.complement256` classes for more information.

    """
    def __init__(self, color, *items):
        (r, g, b) = parse_color(color)
        self.color = xterm256.rgb_to_xterm(r, g, b)
        self.items = items

    def __str__(self):
        return self.fmt % (
            self.color, self.sep.join([unicode(s) for s in self.items]))

class ColorStringTrue(ColorString):
    r"""Base class for 24-bit "truecolor" stylized string-like objects.

    See the :class:`.fgtrue`, :class:`.bgtrue`, :class:`.highlighttrue`, and
    :class:`.complementtrue` classes for more information.

    """
    def __init__(self, color, *items):
        self.color = parse_color(color)
        self.items = items

    def __str__(self):
        return self.fmt % (
            self.color[0], self.color[1], self.color[2], self.sep.join([unicode(s) for s in self.items]))

class plain(ColorString):
    r"""Plain text wrapper

    This class is useful for concatenating plain strings with
    :class:`.ColorString` objects. For example::

        from fabulous.color import plain
        >>> len(plain("hello ", bold("kitty")))
        11

    """
    pass


class bold(ColorString):
    r"""Bold text wrapper

    This class creates a string-like object containing bold or bright text. It
    also brightens the foreground and background colors. This is supported by
    all terminals that support ANSI color codes.

    Example usage::

        from fabulous.color import bold
        print bold('i am bold!')
        print plain('hello ', bold('world'))

    The ANSI escape codes are as follows::

        >>> str(bold("hello"))
        '\x1b[1mhello\x1b[22m'

    """
    fmt = esc(1) + "%s" + esc(22)


class italic(ColorString):
    r"""Italic text wrapper

    This class creates a string-like object containing italic text, which is
    supported by almost no terminals.

    The ANSI escape codes are as follows::

        >>> str(italic("hello"))
        '\x1b[3mhello\x1b[23m'

    """
    fmt = esc(3) + "%s" + esc(23)


class underline(ColorString):
    r"""Underline text wrapper

    This class creates a string-like object containing underline text. This is
    supported by SOME terminals, as documented in the terminal support section.

    Example usage::

        from fabulous.color import underline
        print underline('i am underlined!')
        print plain('hello ', underline('world'))

    The ANSI escape codes are as follows::

        >>> str(underline("hello"))
        '\x1b[4mhello\x1b[24m'

    """
    fmt = esc(4) + "%s" + esc(24)


class underline2(ColorString):
    r"""Alternative underline text wrapper

    See also: :class:`.underline`.

    The ANSI escape codes are as follows::

        >>> str(underline2("hello"))
        '\x1b[21mhello\x1b[24m'

    """
    fmt = esc(21) + "%s" + esc(24)


class strike(ColorString):
    r"""Strike-through text wrapper

    This class creates a string-like object containing strike-through text,
    which is supported by very few terminals.

    Example usage::

        from fabulous.color import strike
        print strike('i am stricken!')
        print plain('hello ', strike('world'))

    The ANSI escape codes are as follows::

        >>> str(strike("hello"))
        '\x1b[9mhello\x1b[29m'

    """
    fmt = esc(9) + "%s" + esc(29)


class blink(ColorString):
    r"""Blinking text wrapper

    This class creates a string-like object containing blinking text. This is
    supported by SOME terminals, as documented in the terminal support section.

    Example usage::

        from fabulous.color import blink
        print blink('i am underlined!')
        print plain('hello ', blink('world'))

    The ANSI escape codes are as follows::

        >>> str(blink("hello"))
        '\x1b[5mhello\x1b[25m'

    """
    fmt = esc(5) + "%s" + esc(25)


class flip(ColorString):
    r"""Flips background and foreground colors

    For example::

        from fabulous.color import flip, red
        print flip(red('hello'))

    Is equivalent to the following on a black terminal::

        from fabulous.color import black, red_bg
        print red_bg(black('hello'))

    The ANSI escape codes are as follows::

        >>> str(flip("hello"))
        '\x1b[7mhello\x1b[27m'

    """
    fmt = esc(7) + "%s" + esc(27)


class black(ColorString):
    r"""Black foreground text wrapper

    This class creates a string-like object containing text with a black
    foreground.

    Example usage::

        from fabulous.color import black
        print black('i am black!')
        print plain('hello ', black('world'))

    Text can be made dark grey by using :class:`.bold`::

        from fabulous.color import bold, black
        print bold(black('i am dark grey!'))

    The ANSI escape codes are as follows::

        >>> str(black("hello"))
        '\x1b[30mhello\x1b[39m'

    """
    fmt = esc(30) + "%s" + esc(39)

class red(ColorString):
    r"""Red foreground text wrapper

    This class creates a string-like object containing text with a red
    foreground.

    Example usage::

        from fabulous.color import red
        print red('i am red!')
        print plain('hello ', red('world'))

    Text can be made bright red by using :class:`.bold`::

        from fabulous.color import bold, red
        print bold(red('i am bright red!'))

    The ANSI escape codes are as follows::

        >>> str(red("hello"))
        '\x1b[31mhello\x1b[39m'

    """
    fmt = esc(31) + "%s" + esc(39)


class green(ColorString):
    r"""Green foreground text wrapper

    This class creates a string-like object containing text with a green
    foreground.

    Example usage::

        from fabulous.color import green
        print green('i am green!')
        print plain('hello ', green('world'))

    Text can be made bright green by using :class:`.bold`::

        from fabulous.color import bold, green
        print bold(green('i am bright green!'))

    The ANSI escape codes are as follows::

        >>> str(green("hello"))
        '\x1b[32mhello\x1b[39m'

    """
    fmt = esc(32) + "%s" + esc(39)


class yellow(ColorString):
    r"""Yellow foreground text wrapper

    This class creates a string-like object containing text with a "yellow"
    foreground, which in many terminals is actually going to look more
    brownish.

    Example usage::

        from fabulous.color import yellow
        print yellow('i am yellow brownish!')
        print plain('hello ', yellow('world'))

    Text can be made true bright yellow by using :class:`.bold`::

        from fabulous.color import bold, yellow
        print bold(yellow('i am bright yellow!'))

    The ANSI escape codes are as follows::

        >>> str(yellow("hello"))
        '\x1b[33mhello\x1b[39m'

    """
    fmt = esc(33) + "%s" + esc(39)


class blue(ColorString):
    r"""Blue foreground text wrapper

    This class creates a string-like object containing text with a blue
    foreground.

    Example usage::

        from fabulous.color import blue
        print blue('i am dark blue!')
        print plain('hello ', blue('world'))

    Text can be made sky blue by using :class:`.bold`::

        from fabulous.color import bold, blue
        print bold(blue('i am sky blue!'))

    The ANSI escape codes are as follows::

        >>> str(blue("hello"))
        '\x1b[34mhello\x1b[39m'

    """
    fmt = esc(34) + "%s" + esc(39)


class magenta(ColorString):
    r"""Purple/magenta foreground text wrapper

    This class creates a string-like object containing text with a magenta
    foreground. Although in many terminals, it's going to look more purple.

    Example usage::

        from fabulous.color import magenta
        print magenta('i am magenta purplish!')
        print plain('hello ', magenta('world'))

    Text can be made hot pink by using :class:`.bold`::

        from fabulous.color import bold, magenta
        print bold(magenta('i am hot pink!'))

    The ANSI escape codes are as follows::

        >>> str(magenta("hello"))
        '\x1b[35mhello\x1b[39m'

    """
    fmt = esc(35) + "%s" + esc(39)


class cyan(ColorString):
    r"""Cyan foreground text wrapper

    This class creates a string-like object containing text with a cyan
    foreground.

    Example usage::

        from fabulous.color import cyan
        print cyan('i am cyan!')
        print plain('hello ', cyan('world'))

    Text can be made bright cyan by using :class:`.bold`::

        from fabulous.color import bold, cyan
        print bold(cyan('i am bright cyan!'))

    The ANSI escape codes are as follows::

        >>> str(cyan("hello"))
        '\x1b[36mhello\x1b[39m'

    """
    fmt = esc(36) + "%s" + esc(39)


class white(ColorString):
    r"""White foreground text wrapper

    This class creates a string-like object containing text with a light grey
    foreground.

    Example usage::

        from fabulous.color import white
        print white('i am light grey!')
        print plain('hello ', white('world'))

    Text can be made true white by using :class:`.bold`::

        from fabulous.color import bold, white
        print bold(white('i am bold white!'))

    The ANSI escape codes are as follows::

        >>> str(white("hello"))
        '\x1b[37mhello\x1b[39m'

    """
    fmt = esc(37) + "%s" + esc(39)


class highlight_black(ColorString):
    r"""Dark grey highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.black`.

    """
    fmt = esc(1, 30, 7) + "%s" + esc(22, 27, 39)


class highlight_red(ColorString):
    r"""Red highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.red`.

    """
    fmt = esc(1, 31, 7) + "%s" + esc(22, 27, 39)


class highlight_green(ColorString):
    r"""Green highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.green`.

    """
    fmt = esc(1, 32, 7) + "%s" + esc(22, 27, 39)


class highlight_yellow(ColorString):
    r"""Yellow highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.yellow`.

    """
    fmt = esc(1, 33, 7) + "%s" + esc(22, 27, 39)


class highlight_blue(ColorString):
    r"""Blue highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.blue`.

    """
    fmt = esc(1, 34, 7) + "%s" + esc(22, 27, 39)


class highlight_magenta(ColorString):
    r"""Hot pink highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.magenta`.

    """
    fmt = esc(1, 35, 7) + "%s" + esc(22, 27, 39)


class highlight_cyan(ColorString):
    r"""Cyan highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.cyan`.

    """
    fmt = esc(1, 36, 7) + "%s" + esc(22, 27, 39)


class highlight_white(ColorString):
    r"""White highlight text wrapper

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.yellow`.

    """
    fmt = esc(1, 37, 7) + "%s" + esc(22, 27, 39)


class black_bg(ColorString):
    r"""Black background text wrapper

    This class creates a string-like object containing text with a black
    background. On properly configured terminals, this will do nothing.

    Example usage::

        from fabulous.color import black_bg
        print black_bg('i have a black background!')
        print plain('hello ', black_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(black_bg("hello"))
        '\x1b[40mhello\x1b[49m'

    """
    fmt = esc(40) + "%s" + esc(49)


class red_bg(ColorString):
    r"""Red background text wrapper

    This class creates a string-like object containing text with a red
    background.

    Example usage::

        from fabulous.color import red_bg
        print red_bg('i have a red background!')
        print plain('hello ', red_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(red_bg("hello"))
        '\x1b[41mhello\x1b[49m'

    """
    fmt = esc(41) + "%s" + esc(49)


class green_bg(ColorString):
    r"""Green background text wrapper

    This class creates a string-like object containing text with a green
    background.

    Example usage::

        from fabulous.color import green_bg
        print green_bg('i have a green background!')
        print plain('hello ', green_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(green_bg("hello"))
        '\x1b[42mhello\x1b[49m'

    """
    fmt = esc(42) + "%s" + esc(49)


class yellow_bg(ColorString):
    r"""Yellow background text wrapper

    This class creates a string-like object containing text with a yellow
    background.

    Example usage::

        from fabulous.color import yellow_bg
        print yellow_bg('i have a yellow background!')
        print plain('hello ', yellow_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(yellow_bg("hello"))
        '\x1b[43mhello\x1b[49m'

    """
    fmt = esc(43) + "%s" + esc(49)


class blue_bg(ColorString):
    r"""Blue background text wrapper

    This class creates a string-like object containing text with a blue
    background.

    Example usage::

        from fabulous.color import blue_bg
        print blue_bg('i have a blue background!')
        print plain('hello ', blue_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(blue_bg("hello"))
        '\x1b[44mhello\x1b[49m'

    """
    fmt = esc(44) + "%s" + esc(49)


class magenta_bg(ColorString):
    r"""Magenta background text wrapper

    This class creates a string-like object containing text with a magenta
    background.

    Example usage::

        from fabulous.color import magenta_bg
        print magenta_bg('i have a magenta background!')
        print plain('hello ', magenta_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(magenta_bg("hello"))
        '\x1b[45mhello\x1b[49m'

    """
    fmt = esc(45) + "%s" + esc(49)


class cyan_bg(ColorString):
    r"""Cyan background text wrapper

    This class creates a string-like object containing text with a cyan
    background.

    Example usage::

        from fabulous.color import cyan_bg
        print cyan_bg('i have a cyan background!')
        print plain('hello ', cyan_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(cyan_bg("hello"))
        '\x1b[46mhello\x1b[49m'

    """
    fmt = esc(46) + "%s" + esc(49)


class white_bg(ColorString):
    r"""White background text wrapper

    This class creates a string-like object containing text with a white
    background.

    Example usage::

        from fabulous.color import white_bg
        print white_bg('i have a white background!')
        print plain('hello ', white_bg('world'))

    The ANSI escape codes are as follows::

        >>> str(white_bg("hello"))
        '\x1b[47mhello\x1b[49m'

    """
    fmt = esc(47) + "%s" + esc(49)


class fg256(ColorString256):
    r"""xterm256 foreground color wrapper

    This class creates a string-like object that has an xterm256 color. The
    color is specified as a CSS color code, which is automatically quantized to
    the available set of xterm colors.

    These colors are more dependable than the 4-bit colors, because 8-bit
    colors don't get changed by the terminal theme. They will consistently be
    the requested color, which is calculated using a simple math formula.

    However it is worth noting that in Terminal.app on Mac OS, 8-bit colors
    appear to be designed rather than formulaic, so they look much nicer.

    Example usage::

        from fabulous import fg256, plain
        print fg256('#F00', 'i am red!')
        print fg256('#FF0000', 'i am red!')
        print fg256('magenta', 'i am', ' magenta!')
        print plain('hello ', fg256('magenta', 'world'))

    The ANSI escape codes look as follows::

        >>> str(fg256('red', 'hello'))
        '\x1b[38;5;196mhello\x1b[39m'

    """
    fmt = esc(38, 5, "%d") + "%s" + esc(39)


class fgtrue(ColorStringTrue):
    r"""24-bit "truecolor" foreground color wrapper

    This class creates a string-like object that has a 24-bit color. The
    color is specified as a CSS color code.

    These colors are, in theory, the most dependable of all... presuming your
    terminal supports them, of course.

    Example usage::

        from fabulous import fgtrue, plain
        print fgtrue('#F00', 'i am red!')
        print fgtrue('#FF0000', 'i am red!')
        print fgtrue('magenta', 'i am', ' magenta!')
        print plain('hello ', fgtrue('magenta', 'world'))

    The ANSI escape codes look as follows::

        >>> str(fgtrue('red', 'hello'))
        '\x1b[38;2;255;0;0mhello\x1b[39m'

    """
    fmt = esc(38, 2, "%d", "%d", "%d") + "%s" + esc(39)


class bg256(ColorString256):
    r"""xterm256 background color wrapper

    This class creates a string-like object that has an xterm256 color. The
    color is specified as a CSS color code, which is automatically quantized to
    the available set of xterm colors.

    These colors are more dependable than the 4-bit colors, because 8-bit
    colors don't get changed by the terminal theme. They will consistently be
    the requested color.

    However it is worth noting that in Terminal.app on Mac OS, 8-bit background
    colors are ever so slightly different than their foreground equivalent.
    Therefore Terminal.app has effectively 512 colors.

    Example usage::

        from fabulous import bg256, plain
        print bg256('#F00', 'i have a red background!')
        print bg256('#FF0000', 'i have a red background!')
        print bg256('magenta', 'i have a', ' magenta background!')
        print plain('hello ', bg256('magenta', 'world'))

    The ANSI escape codes look as follows::

        >>> str(bg256('red', 'hello'))
        '\x1b[48;5;196mhello\x1b[49m'

    """
    fmt = esc(48, 5, "%d") + "%s" + esc(49)


class bgtrue(ColorStringTrue):
    r"""24-bit "truecolor" background color wrapper

    This class creates a string-like object that has a 24-bit color. The
    color is specified as a CSS color code.

    These colors are, in theory, the most dependable of all... presuming your
    terminal supports them, of course.

    Example usage::

        from fabulous import bgtrue, plain
        print bgtrue('#F00', 'i have a red background!')
        print bgtrue('#FF0000', 'i have a red background!')
        print bgtrue('magenta', 'i have a', ' magenta background!')
        print plain('hello ', bgtrue('magenta', 'world'))

    The ANSI escape codes look as follows::

        >>> str(bgtrue('red', 'hello'))
        '\x1b[48;2;255;0;0mhello\x1b[49m'

    """
    fmt = esc(48, 2, "%d", "%d", "%d") + "%s" + esc(49)


class highlight256(ColorString256):
    r"""Highlighted 8-bit color text

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.fg256`.

    """
    fmt = esc(1, 38, 5, "%d", 7) + "%s" + esc(27, 39, 22)


class highlighttrue(ColorStringTrue):
    r"""Highlighted 24-bit "truecolor" color text

    This is equivalent to composing :class:`.bold`, :class:`.flip`, and
    :class:`.fgtrue`.

    """
    fmt = esc(1, 38, 2, "%d", "%d", "%d", 7) + "%s" + esc(27, 39, 22)


class complement256(ColorString256):
    r"""Highlighted 8-bit color text

    This class composes :class:`.bold`, :class:`.flip`, and
    :class:`.bg256`. Then it invokes :meth:`complement` to supply the polar
    opposite :class:`fg256` color.

    This looks kind of hideous at the moment. We're planning on finding a
    better formula for complementary colors in the future.

    """
    fmt = esc(1, 38, 5, "%d", 48, 5, "%d") + "%s" + esc(49, 39, 22)

    def __init__(self, color, *items):
        self.bg = xterm256.rgb_to_xterm(*parse_color(color))
        self.fg = xterm256.rgb_to_xterm(*complement(color))
        self.items = items

    def __str__(self):
        return self.fmt % (
            self.fg, self.bg,
            self.sep.join([unicode(s) for s in self.items]))


class complementtrue(ColorStringTrue):
    r"""Highlighted 24-bit "truecolor" color text

    This class composes :class:`.bold`, :class:`.flip`, and
    :class:`.bgtrue`. Then it invokes :meth:`complement` to supply the polar
    opposite :class:`fgtrue` color.

    This looks kind of hideous at the moment. We're planning on finding a
    better formula for complementary colors in the future.

    """
    fmt = esc(1, 38, 2, "%d", "%d", "%d", 48, 2, "%d", "%d", "%d") + "%s" + esc(49, 39, 22)

    def __init__(self, color, *items):
        self.bg = parse_color(color)
        self.fg = complement(color)
        self.items = items

    def __str__(self):
        return self.fmt % (
            self.fg[0], self.fg[1], self.fg[2],
            self.bg[0], self.bg[1], self.bg[2],
            self.sep.join([unicode(s) for s in self.items]))


def h1(title, line=OVERLINE):
    """Prints bold text with line beneath it spanning width of terminal
    """
    width = utils.term.width
    printy(bold(title.center(width)).as_utf8)
    printy(bold((line * width)[:width]).as_utf8)


def parse_color(color):
    r"""Turns a color into an (r, g, b) tuple

    >>> parse_color('white')
    (255, 255, 255)
    >>> parse_color('#ff0000')
    (255, 0, 0)
    >>> parse_color('#f00')
    (255, 0, 0)
    >>> parse_color((255, 0, 0))
    (255, 0, 0)
    >>> from fabulous import grapefruit
    >>> parse_color(grapefruit.Color((0.0, 1.0, 0.0)))
    (0, 255, 0)
    """
    if isinstance(color, basestring):
        color = grapefruit.Color.NewFromHtml(color)
    if isinstance(color, int):
        (r, g, b) = xterm256.xterm_to_rgb(color)
    elif hasattr(color, 'rgb'):
        (r, g, b) = [int(c * 255.0) for c in color.rgb]
    else:
        (r, g, b) = color
    assert isinstance(r, int) and 0 <= r <= 255
    assert isinstance(g, int) and 0 <= g <= 255
    assert isinstance(b, int) and 0 <= b <= 255
    return (r, g, b)


def complement(color):
    r"""Calculates polar opposite of color

    This isn't guaranteed to look good >_> (especially with brighter, higher
    intensity colors.) This will be replaced with a formula that produces
    better looking colors in the future.

    >>> complement('red')
    (0, 255, 76)
    >>> complement((0, 100, 175))
    (175, 101, 0)

    """
    (r, g, b) = parse_color(color)
    gcolor = grapefruit.Color((r / 255.0, g / 255.0, b / 255.0))
    complement = gcolor.ComplementaryColor()
    (r, g, b) = [int(c * 255.0) for c in complement.rgb]
    return (r, g, b)


def section(title, bar=OVERLINE, strm=sys.stdout):
    """Helper function for testing demo routines
    """
    width = utils.term.width
    printy(bold(title.center(width)))
    printy(bold((bar * width)[:width]))



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import itertools

from fabulous.logs import *
from fabulous.gotham import *
from fabulous.color import *

try:
  next
except NameError:
  next = lambda x: x.next()


logger = logging.getLogger('fabulous')


def luv():
    msg = "hello there how are you?  i love you!  sincerely compy <3 <3"
    while True:
        for n in itertools.chain(range(len(msg)), reversed(range(len(msg)))):
            yield msg[:n+1]


def bad_things():
    yield red("godzilla attack")
    yield bold(red("magnetic poles reverse"))
    yield red("caffeine use criminalized")
    yield bold(red("more horrible things happening...."))
    yield highlight_red("THIS INFORMATION IS NOT BEING DROWNED OUT!")
    yield bold(red("hip hip hooray"))


def test_transientlogger():
    import random, time
    happy, nightmare = luv(), bad_things()
    try:
        while True:
            if random.randrange(60) == 0:
                logger.warning(next(nightmare))
            else:
                logger.debug(next(happy))
            time.sleep(0.02)
    except StopIteration:
        pass


def test_transientlogger2():
    import time, random
    gothic = lorem_gotham()
    try:
        while True:
            if random.randrange(20) == 0:
                logger.warning(red(next(gothic)))
            else:
                logger.debug(next(gothic))
            time.sleep(0.1)
    except StopIteration:
        pass


if __name__ == '__main__':
    basicConfig(level=logging.WARNING)
    logging.warning("RUNNING TEST: test_transientlogger()")
    test_transientlogger()
    logging.warning("RUNNING TEST: test_transientlogger2()")
    test_transientlogger2()



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example values for the answers argument in prompt.query()
"""

true_false = ({'values':(True, 'T', 'Yes','Y',),'fg':'green'},
            {'values':(False, 'F', 'No', 'N'),'fg':'red'})

sex = ({'values':('Male','M'),'fg':'cyan'},
            {'values':('Female','F'),'fg':'magenta'})



# Copyright 2016 The Fabulous Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    fabulous.rlcomplete
    ~~~~~~~~~~~~~~~~~~~

    Readline related stuff.
"""

import os

class Completer(object):
    """A base class for completers.
    
    Child classes should implement the completelist method.
    """
    text = None
    
    delims = '\t\n'
    
    def __init__(self):
        pass
    
    def complete(self, text, state):
        """The actual completion method
        
        This method is not meant to be overridden. Override the
        completelist method instead. It will make your life much easier.
        
        For more detail see documentation for readline.set_completer
        """
        if text != self.text:
            self.matches = self.completelist(text)
            self.text = text
        try:
            return self.matches[state]
        except IndexError:
            return None

    def completelist(self, text):
        """Returns a list.
        
        The list contains a series of strings which are the suggestions
        for the given string ``text``. It is valid to have no suggestions
        (empty list returned).
        """
        return []

class ListCompleter(Completer):
    """A class that does completion based on a predefined list.
    """
    
    def __init__(self, words, ignorecase):
        self.words = words
        self.ignorecase = ignorecase
    
    def completelist(self,text):
        if self.ignorecase:
            return [w for w in self.words if w.lower().startswith(text.lower())]
        else:
            return [w for w in self.words if w.startswith(text)]


class PathCompleter(Completer):
    """Does completion based on file paths. """
    
    def buildpath(self, base, *paths):
        path = os.path.join(base,*paths)
        if os.path.isdir(os.path.expanduser(path)) and path[-1] != os.path.sep:
            path += os.path.sep
        return path
    
    @staticmethod
    def matchuserhome(prefix):
        """To find matches that start with prefix.
        
        For example, if prefix = '~user' this
        returns list of possible matches in form of ['~userspam','~usereggs'] etc.
        
        matchuserdir('~') returns all users
        """
        if not prefix.startswith('~'):
            raise ValueError("prefix must start with ~")
        try: import pwd
        except ImportError:
            try: import winpwd as pwd
            except ImportError: return []
        return ['~' + u[0] for u in pwd.getpwall() if u[0].startswith(prefix[1:])]
        
    
    def completelist(self, text):
        """Return a list of potential matches for completion
        
        n.b. you want to complete to a file in the current working directory
        that starts with a ~, use ./~ when typing in. Paths that start with
        ~ are magical and specify users' home paths
        """
        path = os.path.expanduser(text)
        if len(path) == 0 or path[0] != os.path.sep:
            path = os.path.join(os.getcwd(), path)
        if text == '~':
            dpath = dtext = ''
            bpath = '~'
            files = ['~/']
        elif text.startswith('~') and text.find('/', 1) < 0:
            return self.matchuserhome(text)
        else:
            dtext = os.path.dirname(text)
            dpath = os.path.dirname(path)
            bpath = os.path.basename(path)
            files = os.listdir(dpath)
        if bpath =='':
            matches = [self.buildpath(text, f) for f in files if not f.startswith('.')]
        else:
            matches = [self.buildpath(dtext, f) for f in files if f.startswith(bpath)]
        if len(matches) == 0 and os.path.basename(path)=='..':
            files = os.listdir(path)
            matches = [os.path.join(text, f) for f in files]
        return matches
